/*
Copyright (c) 1997-1998 David Kuehling

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>

enum ARCTYPES
{
  DIVIDED,
  NORMAL,
  SOLID,
  PROGDATA
};

extern void PackArc (unsigned char *InNm[], unsigned Ins, char *OutNm, enum ARCTYPES ArcType);
extern long fsize (FILE *f);
extern char *TempNam (char *Dir, char *Pre);
extern unsigned char HeaderName[256];

unsigned char	*RawFile[256],
		*Archive,
		PrgmComment[256][40],
		FargoVersion;
unsigned	ArcType = NORMAL,
		Files;
long		LengthSum;

void fOpenErr (char *f)
{
  fprintf (stderr, "\n\n\tERROR: can't open file: %s.", f);
  exit (1);
}

void fCreateErr (char *f)
{
  fprintf (stderr, "\n\n\tERROR: can't create file: %s.", f);
  exit (1);
}

void fExistErr (char *f)
{
  fprintf (stderr, "\n\n\tERROR: can't create temporary file: %s. \n\t\tfile allready exists.", f);
  exit (1);
}

void File92pErr (char *f)
{
  fprintf (stderr, "\n\n\tERROR: %s is no real 92p - file.", f);
  exit (1);
}

void UnknComErr (char *c)
{
  fprintf (stderr, "\n\n\tERROR: unknown command: %s", c);
  exit (1);
}

void CompileErr ()
{
  fprintf (stderr, "\n\n\tERROR: can't compile archive with Fargo.");
  exit (1);
}


void PrintHelp ()
{
  printf ("\nSyntax:");
  printf ("\n\tSFX92 [-d] [@]source1 [@]source2 ... archive.92p");
  printf ("\n\t-d\tswitches to divided archive type");
  printf ("\n\t@source\tlist-file - contains source-filenames, divided by linebreaks");
  printf ("\n\tsource\tsourcefile, must be a Fargo program (version 0.1.x or 0.2.x)");
  printf ("\n\tarchive.92p\tname of the self-extracting Fargo program to create\n");
}


unsigned char File92pOk (FILE *f)
{
  fseek (f, 0, SEEK_SET);
  if (fgetc (f) != '*' ||
      fgetc (f) != '*' ||
      fgetc (f) != 'T' ||
      fgetc (f) != 'I' ||
      fgetc (f) != '9' ||
      fgetc (f) != '2' ||
      fgetc (f) != '*' ||
      fgetc (f) != '*')
    return (0);

  fseek (f, 0x58, SEEK_SET);
  if (fgetc (f) ==  0  &&
      fgetc (f) == 'P' &&
      fgetc (f) == '1' &&
      fgetc (f) == '0')
  {
    FargoVersion = 0;
    return (1);
  }
  fseek (f, 0x58, SEEK_SET);
  if (fgetc (f) ==  0  &&
      fgetc (f) == '2' &&
      fgetc (f) == 'E' &&
      fgetc (f) == 'X' &&
      fgetc (f) == 'E' &&
      fgetc (f) == ' ' &&
      fgetc (f) == 'A' &&
      fgetc (f) == 'P' &&
      fgetc (f) == 'P' &&
      fgetc (f) == 'L')
  {
    FargoVersion = 1;
    return (1);
  }
  return (0);
}

char *getPrefix (char *FstFile, char *Prefix)
{
  register int c, I;
  char *pbegin;

  for (pbegin = FstFile, c = *FstFile++; c != '\0'; c = *FstFile++)
    if (c == '\\' || c == ':')
      pbegin = FstFile;

  for (I = 0; I < 5 && *pbegin != '\0' && *pbegin != '.'; I++)
    Prefix[I] = *pbegin++;
  Prefix[I] = '\0';

  return (Prefix);
}


void Convert2Raw (unsigned char *File[])
{
  long	Length;
  register int	I2,
		c;
  unsigned	I;

  unsigned char Prefix[6],
		*RawName;

  FILE *InHnd, *OutHnd;

  for (I = 0, LengthSum = 0; I < Files; I++)
  {
    if ((InHnd = fopen (File[I], "rb")) == NULL)
      fOpenErr (File[I]);

    if (!File92pOk (InHnd))
      File92pErr (File[I]);

    getPrefix (File[I], Prefix);

    if (!(RawFile[I] = TempNam (".", Prefix)) ||
	(OutHnd = fopen (RawFile[I], "wb+")) == NULL)
      fCreateErr ("temporary file");

    printf ("\nConverting %s -> %s", File[I], RawFile[I]);

    /* Copy File Without Header And Checksum*/
    fseek (InHnd, 0x56, SEEK_SET);
    for (I2 = 56, Length = fsize (InHnd)-2;  I2 < Length; I2++)
      fputc (fgetc (InHnd), OutHnd);

    LengthSum += Length - 56;

    /* Get Program Comment */
    for (I2 = 0; I2 < 9; I2++)
      PrgmComment[I][I2] = ' ';

    for (I2 = 0, RawName = File[I];
	 File[I][I2] != '\0'; I2++)
      if (File[I][I2] == '\\' || File[I][I2] == ':')
	RawName = &File[I][I2+1];

    for (I2 = 0; RawName[I2] != '\0' && RawName[I2] != '.'; I2++)
      PrgmComment[I][I2] = toupper (RawName[I2]);

    fseek (OutHnd,
	   FargoVersion == 0 ? 0x0E : 0x16,
	   SEEK_SET);

    fseek (OutHnd, (fgetc (OutHnd) << 8 | fgetc (OutHnd)) +
	   (FargoVersion == 0 ? 2 : 0),
	   SEEK_SET);
    for (I2 = 9, c = fgetc (OutHnd); I2 < 39 && c != '\0';
	 I2++, c = fgetc (OutHnd))
      PrgmComment[I][I2] = c;
    PrgmComment[I][I2] = '\0';

    fclose (InHnd);
    fclose (OutHnd);
  }
}


void CompressRaw (unsigned char *n)
{
  register int I;

  unsigned char Prefix[6];

  getPrefix (n, Prefix);

  printf ("\n\nCompression...");
  if (!(Archive = TempNam (".", Prefix)))
    fCreateErr ("temporary file");

  for (I = 0; Archive[I] != '\0'; I++)
    if (Archive[I] == '/')
      Archive[I] = '\\';

  PackArc (RawFile, Files, Archive, ArcType);
  remove (HeaderName);

  for (I = 0; I < Files; I++)
  {
    remove (RawFile[I]);
    free (RawFile[I]);
  }
}


void CompileArc (unsigned char *n)
{
  static unsigned char	ASMFile[256],
			OnlyName[256],
			FargoCall[256],
			File92p[256],
			EndofName;
  register int I, Bytes;
  long	Length92p;
  FILE	*OutHnd;

  for (EndofName = I = strlen (n) - 1; I >= 0 && n[I] != '\\'; I--)
    if (n[I] == '.')
    {
      EndofName = I;
      break;
    }

  for (I = 0; I < EndofName; I++)
    OnlyName[I] = n[I];
  OnlyName[I] = '\0';

  strcpy (ASMFile, OnlyName);
  strcpy (File92p, OnlyName);
  strcat (ASMFile, ".asm");
  strcat (File92p, ".92p");

  if ((OutHnd = fopen (ASMFile, "rt")) != NULL)
    fExistErr(ASMFile);
  if ((OutHnd = fopen (ASMFile, "wt")) == NULL)
    fCreateErr (ASMFile);

  fprintf (OutHnd, "\tXDEF\t_main\n\tXDEF\t_comment\n");
  fprintf (OutHnd, "\n_main:\n\tlea NameLst(PC),a0");
  fprintf (OutHnd, "\n\tlea Archive(PC),a1");
  fprintf (OutHnd, "\n\tjsr pk92sfx@0000\n\n\trts\n\n");

  for (I = 0, Bytes = 0; I < Files; I++)
  {
    fprintf (OutHnd, "\nName%i:\tdc.b \"%s\",0", I, PrgmComment[I]);
    Bytes += strlen (PrgmComment[I])+1;
  }
  if (Bytes&1)
    fprintf (OutHnd, ",0");

  fprintf (OutHnd, "\n\nNameLst:");
  for (I = 0; I < Files; I++)
    fprintf (OutHnd, "\n\tdc.w Name%i-NameLst", I);
  fprintf (OutHnd, "\n\tdc.w 0");

  fprintf (OutHnd, "\nArchive:\n\tINCBIN \"%s\"", Archive);
  fprintf (OutHnd, "\n\n_comment:\tdc.b \"PK92-SFX - by David Khling\",0");
  fprintf (OutHnd, "\n\n\tEND");
  fclose (OutHnd);

  sprintf (FargoCall, "fargo %s >NUL", OnlyName);
  printf ("\nRunning Fargo...\n");
  if (system (FargoCall))
    CompileErr ();

  remove (Archive);
  free (Archive);

  remove (ASMFile);

  if ((OutHnd = fopen (File92p, "rb")) == NULL)
    fOpenErr ("compiled fargo-program");
  Length92p = fsize (OutHnd);
  fclose (OutHnd);
  printf ("Result: %D >> %D = %6.2f%% (ratio: %6.2f%%)\n", LengthSum, Length92p,
	  (double)(Length92p * 100) / LengthSum,
	  (double)((LengthSum-Length92p) * 100) / LengthSum);
}



int main (unsigned argn, char *argv[])
{
  unsigned	I;
  register int	I2,
		c;
  static unsigned char	FileName[256][128],
			*PtrFlNm[256];
  FILE 	*ListFile;

  printf ("\n\t\t   ---------------------------------------");
  printf ("\n\t\t     SFX92 v1.10 - by David Khling 1998  ");
  printf ("\n\t\t   ---------------------------------------\n");

  printf ("\n\n\t.92p -> Packer92 - selfextracting Fargo program converter\n");

  if (argn < 3)
  {
    PrintHelp ();
    exit (1);
  }

  /* Copy filenames to FileName */
  for (I = 1, Files = 0, PtrFlNm[0] = FileName[0]; I < argn-1; I++)
  {
    if (argv[I][0] == '@')
    {
      if ((ListFile = fopen (&argv[I][1], "rt")) == NULL)
      {
	fprintf (stderr, "\n\n\tERROR: can't open list-file: %s\n", &argv[I][1]);
	return (1);
      }
      for (I2 = 0;;)
      {
	if ((c = fgetc (ListFile)) == EOF)
	{
	  if (I2)
	    FileName[Files++][I2] = '\0';
	  break;
	}
	if (c == '\n')
	{
	  if (I2)
	    FileName[Files++][I2] = '\0';
	  if (Files >= 255)
	  {
	    fprintf (stderr, "\n\n\tERROR: More than 255 sourcefiles.\n");
	    return (1);
	  }
	  I2 = 0;
	  PtrFlNm[Files] = FileName[Files];
	  continue;
	}
	if (I2 >= 127)
	{
	  fprintf (stderr, "\n\n\tERROR: Too long filename in list-file %s.\n", &argv[I][1]);
	  return (1);
	}
	FileName[Files][I2++] = c;
      }
      fclose (ListFile);
    }
    else if (argv[I][0] == '-' || argv[I][0] == '/')
    {
      if (!stricmp (&argv[I][0], "-d"))
	ArcType = DIVIDED;
      else if (!stricmp (&argv[I][0], "n"))
	ArcType = NORMAL;
      else
	UnknComErr (argv[I]);
    }
    else
    {
      PtrFlNm[Files] = FileName[Files];
      strcpy (FileName[Files++], argv[I]);
    }
  }
  if (!Files)
  {
    fprintf (stderr, "\n\n\tERROR: No sourcefiles.\n");
    return (1);
  }

  Convert2Raw (PtrFlNm);
  CompressRaw (argv[argn-1]);
  CompileArc (argv[argn-1]);

  return (0);
}
